import os
import sys
import typing

import asn1tools
import wx
import wx.svg

import asn1editor
from asn1editor.ASN1SpecHandler import ASN1SpecHandler
from asn1editor.Plugin import Plugin
from asn1editor.PluginInterface import PluginInterface
from asn1editor.wxPython import Environment
from asn1editor.wxPython import WxPythonViewFactory
from asn1editor.wxPython.FilePickerHandler import FilePickerHandler
from asn1editor.wxPython.Resources import resource_path
from asn1editor.wxPython.Styler import Styler


class MainWindow(wx.Frame, PluginInterface):
    def __init__(self, plugins: typing.Optional[typing.List[Plugin]] = None, title=f'ASN.1 editor {asn1editor.__version__}'):
        super(MainWindow, self).__init__(None, title=title, size=(500, 800))

        Environment.load()

        self.__plugins = plugins
        if self.__plugins is not None:
            for plugin in self.__plugins:
                plugin.connect(self)

        self.SetSize(wx.Size(Environment.settings.get('size', (500, 800))))
        self.Maximize(Environment.settings.get('maximized', True))
        self.SetPosition(wx.Point(Environment.settings.get('position', (0, 0))))

        self.__main_panel = wx.ScrolledWindow(self, style=wx.HSCROLL | wx.VSCROLL)
        self.__main_panel.SetScrollbars(15, 15, 50, 50)
        sizer = wx.BoxSizer(wx.VERTICAL)
        self.__main_panel.SetSizer(sizer)
        self.__main_panel.SetAutoLayout(True)

        self._status_bar = self.CreateStatusBar()

        self.__create_menu()
        self.__asn1_handler = None

        self.__model = None
        self.__view = None
        self.__controller = None
        self.__type_name = None
        self.__file_name = None

        self.__progress_window: typing.Optional[wx.ProgressDialog] = None

        self.bind_events()

        # noinspection SpellCheckingInspection
        sys.excepthook = self.__exception_handler

    def __create_menu(self):
        menu_bar = wx.MenuBar()
        file_menu = wx.Menu()
        self.__load_spec_item: wx.MenuItem = file_menu.Append(wx.ID_ANY, 'Open ASN.1 specification')
        # noinspection PyArgumentList
        image: wx.svg.SVGimage = wx.svg.SVGimage.CreateFromFile(resource_path('icons/open.svg'))
        self.__load_spec_item.SetBitmap(image.ConvertToBitmap(width=16, height=16))
        file_menu.AppendSeparator()
        self.__load_data_item: wx.MenuItem = file_menu.Append(wx.ID_OPEN, 'Load encoded data')
        # noinspection PyArgumentList
        image: wx.svg.SVGimage = wx.svg.SVGimage.CreateFromFile(resource_path('icons/load_encoded.svg'))
        self.__load_data_item.SetBitmap(image.ConvertToBitmap(width=16, height=16))
        self.__load_data_item.Enable(False)
        self.__save_data_item: wx.MenuItem = file_menu.Append(wx.ID_SAVE, 'Save encoded data')
        # noinspection PyArgumentList
        image: wx.svg.SVGimage = wx.svg.SVGimage.CreateFromFile(resource_path('icons/save_encoded.svg'))
        self.__save_data_item.SetBitmap(image.ConvertToBitmap(width=16, height=16))
        self.__save_data_item.Enable(False)
        file_menu.AppendSeparator()
        self.__exit_item = file_menu.Append(wx.ID_EXIT, 'Exit', 'Exit application')
        # noinspection PyArgumentList
        image: wx.svg.SVGimage = wx.svg.SVGimage.CreateFromFile(resource_path('icons/exit.svg'))
        self.__exit_item.SetBitmap(image.ConvertToBitmap(width=16, height=16))
        menu_bar.Append(file_menu, '&File')

        if self.__plugins is not None:

            for plugin_index, plugin in enumerate(self.__plugins):
                plugin_menu = wx.Menu()
                menus = plugin.get_menus()
                for i, menu in enumerate(menus):
                    if not len(menu[0]):
                        plugin_menu.AppendSeparator()
                    else:
                        menu_item: wx.MenuItem = plugin_menu.Append(plugin_index * 1000 + i, menu[0])
                        if menu[1] is not None:
                            self.Bind(wx.EVT_MENU, self.__plugin_menu_event, menu_item)
                        else:
                            menu_item.Enable(False)

                menu_bar.Append(plugin_menu, plugin.get_name())

        help_menu = wx.Menu()
        about_item = help_menu.Append(wx.ID_ABOUT, 'About')
        self.Bind(wx.EVT_MENU, self.__about_item_event, about_item)
        menu_bar.Append(help_menu, '&Help')

        self.SetMenuBar(menu_bar)

    def __plugin_menu_event(self, e):
        plugin_index = e.GetId() // 1000
        self.__plugins[plugin_index].get_menus()[e.GetId() % 1000][1]()

    def bind_events(self):
        def schema_dialog_constructor() -> wx.FileDialog:
            return wx.FileDialog(self, "ASN.1 schema", wildcard="ASN.1 files (*.asn)|*.asn",
                                 style=wx.FD_OPEN | wx.FD_FILE_MUST_EXIST | wx.FD_MULTIPLE)

        def data_load_dialog_constructor() -> wx.FileDialog:
            extensions = ';'.join(ASN1SpecHandler.get_extensions())
            return wx.FileDialog(self, "ASN.1 encoded file", wildcard=f"ASN.1 encoded ({extensions})|{extensions}",
                                 style=wx.FD_OPEN | wx.FD_FILE_MUST_EXIST)

        def data_save_dialog_constructor() -> wx.FileDialog:
            extensions = ';'.join(ASN1SpecHandler.get_extensions())
            return wx.FileDialog(self, "ASN.1 encoded file", wildcard=f"ASN.1 encoded ({extensions})|{extensions}", style=wx.FD_SAVE)

        self.Bind(wx.EVT_MENU, self.exit, self.__exit_item)

        picker = FilePickerHandler(schema_dialog_constructor, self.load_spec)
        self.Bind(wx.EVT_MENU, picker.on_menu_click, self.__load_spec_item)

        picker = FilePickerHandler(data_load_dialog_constructor, self.load_data_from_file)
        self.Bind(wx.EVT_MENU, picker.on_menu_click, self.__load_data_item)

        picker = FilePickerHandler(data_save_dialog_constructor, self.save_data_to_file, True)
        self.Bind(wx.EVT_MENU, picker.on_menu_click, self.__save_data_item)

        self.Bind(wx.EVT_CLOSE, self.close, self)

    def load_spec(self, file_name: str, type_name: typing.Optional[str] = None):
        # Spec file loaded, compile it to show a selection of type names
        if not self.__asn1_handler or file_name not in self.__asn1_handler.get_filename():
            self.__asn1_handler = ASN1SpecHandler(file_name)

        if type_name is None:
            types = self.__asn1_handler.get_types()
            dialog = wx.SingleChoiceDialog(self, 'Select type from ASN.1 file', 'Select type', types)
            try:
                if dialog.ShowModal() == wx.ID_OK:
                    self.__type_name = dialog.GetStringSelection()
            finally:
                dialog.Destroy()
        else:
            self.__type_name = type_name
        if self.__type_name is not None:
            self._status_bar.SetStatusText(f'Loaded {file_name}')
            self.__file_name = file_name

            styler = Styler(os.path.splitext(file_name)[0] + '.style')
            view_factory = WxPythonViewFactory.WxPythonViewFactory(self.__main_panel, styler)
            view_factory.freeze()

            if self.__view is not None:
                self.__view.destroy()

            self.__view, self.__controller = self.__asn1_handler.create_mvc_for_type(self.__type_name, view_factory)
            sizer: wx.Sizer = self.__main_panel.GetSizer()
            sizer.Clear()
            sizer.Add(self.__view.realize(), 0, wx.ALL | wx.EXPAND, 5)
            sizer.Layout()

            self.__main_panel.SetSizer(sizer)
            self.__main_panel.FitInside()
            self.__main_panel.AdjustScrollbars()

            view_factory.thaw()

            self.__load_data_item.Enable(True)
            self.__save_data_item.Enable(True)

    def load_data_from_file(self, file_name: str):
        self.__controller.model_to_view(self.__asn1_handler.load_data_file(file_name))
        self._status_bar.SetStatusText(f'Loaded {file_name} for {self.__type_name}')

    def save_data_to_file(self, file_name: str):
        self.__asn1_handler.save_data_file(file_name, self.__controller.view_to_model())

    def show_data(self, data: bytes, codec: str):
        self.__controller.model_to_view(self.__asn1_handler.get_model_from_data(data, codec))
        self._status_bar.SetStatusText(f'Loaded data for {self.__type_name}')

    def file_picker(self, message: str, wildcard: str, open_: bool) -> typing.Optional[str]:
        def dialog_constructor() -> wx.FileDialog:
            return wx.FileDialog(self, message, wildcard=wildcard,
                                 style=(wx.FD_OPEN | wx.FD_FILE_MUST_EXIST) if open_ else wx.FD_SAVE)

        picker = FilePickerHandler(dialog_constructor, None, not open_)
        picker.on_menu_click(wx.EVT_MENU)
        return picker.filename

    def dir_picker(self, message: str) -> typing.Optional[str]:
        def dialog_constructor() -> wx.DirDialog:
            return wx.DirDialog(self, message,
                                style=wx.DD_DEFAULT_STYLE | wx.DD_DIR_MUST_EXIST)

        picker = FilePickerHandler(dialog_constructor, None)
        picker.on_menu_click(wx.EVT_MENU)
        return picker.filename

    def encode_data(self, codec: str) -> bytes:
        return self.__asn1_handler.get_data_from_model(self.__controller.view_to_model(), codec)

    def get_spec_filename(self) -> str:
        return self.__file_name

    def get_typename(self) -> str:
        return self.__type_name

    def get_spec(self, codec: str):
        return self.__asn1_handler.get_compiled(codec)

    def text_entry(self, message: str, default: typing.Optional[str] = None) -> typing.Optional[str]:
        with wx.TextEntryDialog(self, message) as text_dialog:
            if default is not None:
                text_dialog.SetValue(default)
            if text_dialog.ShowModal() == wx.ID_CANCEL:
                return
            return text_dialog.GetValue()

    def choice_entry(self, message: str, caption: str, choices: typing.List[str], default: typing.Optional[str] = None) -> typing.Optional[str]:
        with wx.SingleChoiceDialog(self, message, caption, choices=choices) as choice_dialog:
            if default is not None:
                try:
                    choice_dialog.SetSelection(choices.index(default))
                except ValueError:
                    pass
            if choice_dialog.ShowModal() == wx.ID_CANCEL:
                return
            return choice_dialog.GetStringSelection()

    def show_status(self, message: str):
        self._status_bar.SetStatusText(message)

    def show_message(self, message: str, caption: str, message_type: PluginInterface.MessageType) -> bool:
        style = wx.CENTER | {PluginInterface.MessageType.WARNING: wx.OK | wx.ICON_WARNING,
                             PluginInterface.MessageType.INFO: wx.OK | wx.ICON_INFORMATION,
                             PluginInterface.MessageType.ERROR: wx.OK | wx.ICON_ERROR,
                             PluginInterface.MessageType.QUESTION: wx.YES_NO | wx.ICON_QUESTION}[message_type]
        ret = wx.MessageBox(message, caption, style=style)
        return True if message_type != PluginInterface.MessageType.QUESTION else wx.YES == ret

    def show_progress(self, message: str, caption: str, max_progress: typing.Optional[int] = None):
        self.__progress_window = wx.ProgressDialog(caption, message, maximum=max_progress if max_progress else 100, style=wx.PD_APP_MODAL | wx.PD_CAN_ABORT)

    def update_progress(self, message: typing.Optional[str] = None, close: bool = False, progress: typing.Optional[int] = None) -> bool:
        running = False
        if self.__progress_window:
            if progress is not None:
                running = self.__progress_window.Update(progress, newmsg=message)[0]
            else:
                running = self.__progress_window.Pulse()[0]
        if close:
            self.__progress_window.Close()
            self.__progress_window = None

        return running

    def get_settings(self) -> dict:
        return Environment.settings.setdefault('Plugin', {})

    def __exception_handler(self, exc_type, value, trace):
        import traceback
        trace = ''.join(traceback.format_exception(exc_type, value, trace))
        print(trace)

        message = f'{value}\n\n{str(exc_type)}'
        exception_str = f'{message}\n\n{trace}'
        wx.MessageBox(message, 'Error', wx.OK | wx.ICON_ERROR, parent=self)
        self._status_bar.SetStatusText(f'Error: {value}')
        Environment.log_error(exception_str)

    # noinspection PyUnusedLocal
    def __about_item_event(self, e):
        del e
        dialog = wx.MessageDialog(self, f'''asn1editor {asn1editor.__version__}

Published under MIT License

Copyright (c) 2020 Florian Fetz
https://github.com/Futsch1/asn1editor

Based on eerimoq's asn1tools, used in {asn1tools.version.__version__}
https://github.com/eerimoq/asn1tools
''', style=wx.ICON_INFORMATION | wx.OK, caption='About')
        dialog.ShowModal()

    # noinspection PyUnusedLocal
    def exit(self, e: wx.Event):
        del e
        self.Close()

    # noinspection PyUnusedLocal
    def close(self, e: wx.Event):
        Environment.settings['size'] = (self.GetSize().Get())
        Environment.settings['maximized'] = self.IsMaximized()
        Environment.settings['position'] = self.GetPosition().Get()

        Environment.save()

        self.Destroy()



from typing import List, Tuple, Optional, Union

import wx
import wx.lib.masked.numctrl
import wx.svg

from asn1editor.interfaces.BitstringInterface import BitstringInterface
from asn1editor.interfaces.OptionalInterface import OptionalInterface
from asn1editor.interfaces.ValueInterface import ValueInterface
from asn1editor.view.AbstractView import AbstractView, ContainerView, ListView, ChoiceView
from asn1editor.view.AbstractViewFactory import AbstractViewFactory
from asn1editor.wxPython.Resources import resource_path
from asn1editor.wxPython.Styler import Styler
from asn1editor.wxPython.WxPythonViews import WxPythonValueView, WxPythonView, WxPythonContainerView, WxPythonListView, WxPythonBooleanView, \
    WxPythonChoiceView, WxPythonBitstringView, WxPythonHexStringView, WxPythonValueSelectionView


class WxPythonViewFactory(AbstractViewFactory):
    def __init__(self, window: wx.ScrolledWindow, styler: Styler):
        self._window = window
        self._styler = styler

    def get_enumerated_view(self, name: str, choices: List[str], optional: bool) -> Tuple[AbstractView, ValueInterface, OptionalInterface]:
        sizer = wx.BoxSizer(wx.HORIZONTAL)
        optional_control = self._add_name_control(sizer, name, optional, ':', 'enumerated')
        edit = wx.Choice(self._window, choices=choices)
        sizer.Add(edit, proportion=1, flag=wx.ALL | wx.EXPAND, border=5)

        style = self._styler.get_style(name)
        if style == 'read_only':
            edit.Enabled(False)
        if style == 'hidden':
            sizer.ShowItems(False)

        view = WxPythonValueSelectionView(sizer, edit, optional_control)
        return view, view, view if optional else None

    def get_text_view(self, name: str, text: str) -> AbstractView:
        sizer = wx.BoxSizer(wx.HORIZONTAL)
        dummy_text = wx.StaticText(self._window, wx.ID_ANY, text)
        self._add_name_control(sizer, name, False)
        sizer.Add(dummy_text, flag=wx.ALL, border=5)

        view = WxPythonView(sizer, None)

        return view

    def get_container_view(self, name: str, optional: bool) -> Tuple[ContainerView, OptionalInterface]:
        sizer = wx.StaticBoxSizer(wx.VERTICAL, self._window, name)
        sizer.Add(self._get_svg('sequence'))
        if optional:
            optional_control = self._add_name_control(sizer, name, optional)
        else:
            optional_control = None

        container_sizer = wx.FlexGridSizer(cols=2, vgap=8, hgap=8)
        sizer.Add(container_sizer)

        style = self._styler.get_style(name)
        if style == 'hidden':
            sizer.ShowItems(False)

        view = WxPythonContainerView(sizer, container_sizer, optional_control)

        return view, view if optional else None

    def get_list_view(self, name: str, minimum: int, maximum: int, optional: bool) -> Tuple[ListView, ValueInterface, OptionalInterface]:
        sizer = wx.StaticBoxSizer(wx.VERTICAL, self._window, name)
        sizer.Add(self._get_svg('sequence_of'))
        if optional:
            optional_control = self._add_name_control(sizer, name, optional)
        else:
            optional_control = None

        num_elements_sizer = wx.BoxSizer(wx.HORIZONTAL)
        num_elements_label = wx.StaticText(self._window, wx.ID_ANY, "Elements:")
        num_elements_sizer.Add(num_elements_label, flag=wx.ALL | wx.ALIGN_CENTER_VERTICAL, border=5)
        num_elements = wx.SpinCtrl(self._window)
        if minimum is not None:
            num_elements.SetMin(minimum)
        else:
            minimum = 0
        if maximum is not None:
            num_elements.SetMax(maximum)
        else:
            maximum = 'infinite'
        num_elements.SetToolTip(f"Minimum elements: {minimum}, maximum elements: {maximum}")
        num_elements_sizer.Add(num_elements, flag=wx.ALL | wx.ALIGN_CENTER_VERTICAL, border=5)

        content = wx.BoxSizer(wx.VERTICAL)
        content.Add(num_elements_sizer)

        sizer.Add(content)

        style = self._styler.get_style(name)
        if style == 'read_only':
            num_elements.Enabled(False)
        if style == 'hidden':
            sizer.ShowItems(False)

        view = WxPythonListView(sizer, num_elements, optional_control)

        return view, view, view if optional else None

    def get_number_view(self, name: str, optional: bool, minimum: Optional[Union[int, float]],
                        maximum: Optional[Union[int, float]], float_: bool) -> Tuple[AbstractView, ValueInterface, OptionalInterface]:
        sizer = wx.BoxSizer(wx.HORIZONTAL)

        optional_control = self._add_name_control(sizer, name, optional, ':', 'float' if float_ else 'integer')
        edit = wx.lib.masked.numctrl.NumCtrl(self._window)
        tool_tip = []
        if isinstance(minimum, int) or isinstance(minimum, float):
            edit.SetAllowNegative(minimum < 0)
            edit.SetMin(minimum)
            tool_tip.append(f'Minimum: {minimum}')
        if isinstance(maximum, int) or isinstance(maximum, float):
            edit.SetMax(maximum)
            tool_tip.append(f'Maximum: {maximum}')
        if float_:
            edit.SetFractionWidth(6)
        if len(tool_tip):
            edit.SetToolTip(', '.join(tool_tip))

        sizer.Add(edit, proportion=1, flag=wx.ALL | wx.EXPAND, border=5)

        style = self._styler.get_style(name)
        if style == 'read_only':
            edit.SetEditable(False)
        if style == 'hidden':
            sizer.ShowItems(False)

        view = WxPythonValueView(sizer, edit, optional_control)
        return view, view, view if optional else None

    def get_boolean_view(self, name: str, optional: bool) -> Tuple[AbstractView, ValueInterface, OptionalInterface]:
        sizer = wx.BoxSizer(wx.HORIZONTAL)
        optional_control = self._add_name_control(sizer, name, optional, ':', 'bool')
        check_box = wx.CheckBox(self._window)
        sizer.Add(check_box, proportion=1, flag=wx.ALL | wx.EXPAND, border=5)

        style = self._styler.get_style(name)
        if style == 'read_only':
            check_box.Enable(False)
        if style == 'hidden':
            sizer.ShowItems(False)

        view = WxPythonBooleanView(sizer, check_box, optional_control)
        return view, view, view if optional else None

    def get_string_view(self, name: str, string_type: str, optional: bool, minimum: Optional[int], maximum: Optional[int]):
        sizer = wx.BoxSizer(wx.HORIZONTAL)
        optional_control = self._add_name_control(sizer, name, optional, ':', 'string', string_type)
        edit = wx.TextCtrl(self._window)
        if maximum:
            edit.SetMaxLength(maximum)
        else:
            maximum = 'infinite'
        if minimum is None:
            minimum = '0'
        edit.SetToolTip(f"Minimum characters: {minimum}, maximum characters: {maximum}")

        sizer.Add(edit, proportion=1, flag=wx.ALL | wx.EXPAND, border=5)

        style = self._styler.get_style(name)
        if style == 'read_only':
            edit.SetEditable(False)
        if style == 'hidden':
            sizer.ShowItems(False)

        view = WxPythonValueView(sizer, edit, optional_control)
        return view, view, view if optional else None

    def get_hex_string_view(self, name: str, optional: bool, minimum: Optional[int], maximum: Optional[int]):
        sizer = wx.BoxSizer(wx.HORIZONTAL)
        optional_control = self._add_name_control(sizer, name, optional, ':', 'string', 'OCTET STRING')
        selector = wx.CheckBox(self._window, label='Hex')
        sizer.Add(selector, flag=wx.ALL | wx.ALIGN_CENTER_VERTICAL, border=5)
        edit = wx.TextCtrl(self._window)

        sizer.Add(edit, proportion=1, flag=wx.ALL | wx.EXPAND, border=5)

        style = self._styler.get_style(name)
        if style == 'read_only':
            edit.SetEditable(False)
        if style == 'hidden':
            sizer.ShowItems(False)

        view = WxPythonHexStringView(sizer, edit, selector, minimum, maximum, optional_control)
        return view, view, view if optional else None

    def get_choice_view(self, name: str, choices: List[str], optional: bool) -> Tuple[ChoiceView, ValueInterface, OptionalInterface]:
        sizer = wx.BoxSizer(wx.HORIZONTAL)
        optional_control = self._add_name_control(sizer, name, optional, icon='choice')

        choice_element_sizer = wx.BoxSizer(wx.HORIZONTAL)
        choice_element_label = wx.StaticText(self._window, wx.ID_ANY)
        choice_element_sizer.Add(choice_element_label, flag=wx.ALL | wx.ALIGN_CENTER_VERTICAL, border=5)
        choice_element = wx.Choice(self._window, choices=choices)
        choice_element_sizer.Add(choice_element, flag=wx.ALL | wx.ALIGN_CENTER_VERTICAL, border=5)

        content = wx.BoxSizer(wx.VERTICAL)
        content.Add(choice_element_sizer)

        sizer.Add(content, flag=wx.ALL | wx.ALIGN_CENTER_VERTICAL)

        style = self._styler.get_style(name)
        if style == 'read_only':
            choice_element.Enable(False)
        if style == 'hidden':
            sizer.ShowItems(False)

        view = WxPythonChoiceView(sizer, choice_element, optional_control)

        return view, view, view if optional else None

    def get_bitstring_view(self, name: str, number_of_bits: int, named_bits: List[Tuple[str, int]], optional: bool) -> Tuple[AbstractView, BitstringInterface,
                                                                                                                             OptionalInterface]:
        sizer = wx.BoxSizer(wx.HORIZONTAL)
        optional_control = self._add_name_control(sizer, name, optional, icon='bitstring')

        checkboxes: List[Tuple[int, wx.CheckBox]] = []

        style = self._styler.get_style(name)

        bits_sizer = wx.StaticBoxSizer(wx.VERTICAL, self._window, "Bits")
        if named_bits:
            for name, bit in named_bits:
                bit_checkbox = wx.CheckBox(self._window, label=f"{bit}: {name}")
                bits_sizer.Add(bit_checkbox)
                if style == 'read_only':
                    bit_checkbox.Enable(False)
                checkboxes.append((bit, bit_checkbox))
        else:
            for bit in range(number_of_bits):
                bit_checkbox = wx.CheckBox(self._window, label=str(bit))
                bits_sizer.Add(bit_checkbox)
                if style == 'read_only':
                    bit_checkbox.Enable(False)
                checkboxes.append((bit, bit_checkbox))

        sizer.Add(bits_sizer)

        if style == 'hidden':
            sizer.ShowItems(False)

        view = WxPythonBitstringView(sizer, checkboxes, optional_control)

        return view, view, view if optional else None

    def update(self):
        self._window.Layout()
        self._window.FitInside()
        self._window.AdjustScrollbars()

    def freeze(self):
        self._window.Freeze()

    def thaw(self):
        self._window.Thaw()

    def _add_name_control(self, sizer: wx.BoxSizer, name: str, optional: bool, suffix: str = '', icon: str = None, icon_tooltip: str = None) -> \
            Optional[wx.CheckBox]:
        flags = wx.ALL | (0 if sizer.GetOrientation() == wx.VERTICAL else wx.ALIGN_CENTER_VERTICAL)
        if optional:
            control = wx.CheckBox(self._window, wx.ID_ANY, name + suffix)
            control.SetToolTip("Optional element")
        else:
            control = wx.StaticText(self._window, wx.ID_ANY, name + suffix)
        if icon is not None:
            sizer.Add(self._get_svg(icon, icon_tooltip), flag=flags)
        sizer.Add(control, flag=flags, border=5)
        return control if optional else None

    def _get_svg(self, bitmap_name: str, icon_tooltip: str = None) -> wx.StaticBitmap:
        # noinspection PyArgumentList
        image: wx.svg.SVGimage = wx.svg.SVGimage.CreateFromFile(resource_path(f'icons/{bitmap_name}.svg'))
        bitmap = wx.StaticBitmap(self._window, bitmap=image.ConvertToBitmap(width=16, height=16))
        if icon_tooltip:
            bitmap.SetToolTip(icon_tooltip)
        else:
            bitmap.SetToolTip(bitmap_name.upper().replace('_', ' '))
        return bitmap



import os
import sys


def resource_path(relative_path: str) -> str:
    """ Get absolute path to resource required for PyInstaller """
    # noinspection SpellCheckingInspection
    base_path = getattr(sys, '_MEIPASS', os.path.dirname(os.path.abspath(__file__)))
    return os.path.join(base_path, relative_path)



import typing
from typing import Optional, Callable, List, Tuple

import wx

from asn1editor.interfaces.BitstringInterface import BitstringInterface
from asn1editor.interfaces.OptionalInterface import OptionalInterface
from asn1editor.interfaces.ValueInterface import ValueInterface
from asn1editor.view.AbstractView import AbstractView, ContainerView, ListView, ChoiceView


class WxPythonView(AbstractView, OptionalInterface):
    def __init__(self, sizer: wx.Sizer, optional_control: Optional[wx.CheckBox]):
        self._sizer = sizer
        self._optional_control = optional_control

    def register_optional_event(self, callback: Callable):
        # noinspection PyUnusedLocal
        def event_closure(e: wx.Event):
            del e
            callback()
            self.enable(self.get_has_value())

        if self._optional_control is not None:
            self._optional_control.Bind(wx.EVT_CHECKBOX, event_closure)

    def register_change_event(self, callback: Callable):
        pass

    def get_has_value(self) -> bool:
        if self._optional_control:
            return self._optional_control.GetValue()
        else:
            return True

    def set_has_value(self, val: bool):
        if self._optional_control:
            self._optional_control.SetValue(val)
            self.enable(val)

    def realize(self) -> wx.Sizer:
        return self._sizer

    def enable(self, enabled: bool):
        return

    def destroy(self):
        self._sizer.Clear(True)


class WxPythonValueView(WxPythonView, ValueInterface):
    def __init__(self, sizer: wx.Sizer, value_control: Optional[wx.TextCtrl] = None,
                 optional_control: Optional[wx.CheckBox] = None):
        super(WxPythonValueView, self).__init__(sizer, optional_control)
        self._value_control = value_control

    def register_change_event(self, callback: Callable):
        # noinspection PyUnusedLocal
        def event_closure(e: wx.Event):
            del e
            self.__update_tooltip()
            callback()

        if self._value_control is not None:
            self._value_control.Bind(wx.EVT_TEXT, event_closure)

    def get_value(self) -> str:
        return self._value_control.GetValue()

    def set_value(self, val: str):
        self._value_control.SetValue(val)
        self.__update_tooltip()

    def enable(self, enabled: bool):
        self._value_control.Enable(enabled)

    def __update_tooltip(self):
        if isinstance(self.get_value(), str) and len(self.get_value()) > 10:
            tool_tip = self._value_control.GetToolTip()
            if tool_tip is not None:
                previous_tooltip = self._value_control.GetToolTip().GetTip().split('\n')
                previous_tooltip = previous_tooltip[-1]
            else:
                previous_tooltip = ''
            self._value_control.SetToolTip('\n'.join([self.get_value(), previous_tooltip]))


class WxPythonValueSelectionView(WxPythonView, ValueInterface):
    def __init__(self, sizer: wx.Sizer, value_control: Optional[wx.Choice] = None,
                 optional_control: Optional[wx.CheckBox] = None):
        super(WxPythonValueSelectionView, self).__init__(sizer, optional_control)
        self._value_control = value_control

    def register_change_event(self, callback: Callable):
        # noinspection PyUnusedLocal
        def event_closure(e: wx.Event):
            del e
            callback()

        if self._value_control is not None:
            self._value_control.Bind(wx.EVT_CHOICE, event_closure)

    def get_value(self) -> str:
        return self._value_control.GetStringSelection()

    def set_value(self, val: str):
        self._value_control.SetStringSelection(val)

    def enable(self, enabled: bool):
        self._value_control.Enable(enabled)


class WxPythonHexStringView(WxPythonView, ValueInterface):
    def __init__(self, sizer: wx.Sizer, value_control: wx.TextCtrl, hex_selector: wx.CheckBox, minimum: Optional[int], maximum: Optional[int],
                 optional_control: Optional[wx.CheckBox] = None):
        super(WxPythonHexStringView, self).__init__(sizer, optional_control)

        self._value_control = value_control
        self._real_value = b''
        self._hex_selector = hex_selector
        self._hex_selector.Bind(wx.EVT_CHECKBOX, self.hex_selector_changed)
        self._hex = self._is_hex()
        self._minimum = minimum
        if self._minimum is None:
            self._minimum = 0
        self._maximum = maximum
        self._update_length()

    def register_change_event(self, callback: Callable):
        # noinspection PyUnusedLocal
        def event_closure(e: wx.Event):
            del e
            self.text_changed()
            callback()

        if self._value_control is not None:
            self._value_control.Bind(wx.EVT_TEXT, event_closure)

    # noinspection PyUnusedLocal
    def hex_selector_changed(self, e: wx.CommandEvent):
        self._hex_selector_changed()

    def text_changed(self):
        if self._hex:
            try:
                self._real_value = bytes.fromhex(self._value_control.GetValue())
            except ValueError:
                pass
        else:
            self._real_value = self._value_control.GetValue().encode('latin-1')

    def _update_control(self):
        if self._hex:
            val = self._real_value.hex()
            val = ' '.join(val[i:i + 2] for i in range(0, len(val), 2))
        else:
            val = self._real_value.decode('latin-1', errors='replace')

        self._value_control.ChangeValue(val)

        self.__update_tooltip()

    def _hex_selector_changed(self):
        if self._hex != self._is_hex():
            if not self._hex:
                # Was ASCII before, now is hex
                if self._maximum:
                    self._maximum *= 2
                self._minimum *= 2
            else:
                if self._maximum:
                    self._maximum //= 2
                self._minimum //= 2

            self._update_length()
            self._hex = self._is_hex()
            self._update_control()

    def __update_tooltip(self):
        if len(self.get_value()) > 10:
            tool_tip = self._value_control.GetToolTip()
            if tool_tip is not None:
                previous_tooltip = self._value_control.GetToolTip().GetTip().split('\n')
                previous_tooltip = previous_tooltip[-1]
            else:
                previous_tooltip = ''
            self._value_control.SetToolTip('\n'.join([str(self.get_value()), previous_tooltip]))

    def get_value(self) -> bytes:
        return self._real_value

    def set_value(self, val: bytes):
        assert isinstance(val, bytes)
        self._real_value = val
        self._update_control()

    def enable(self, enabled: bool):
        self._value_control.Enable(enabled)

    def _is_hex(self) -> bool:
        return self._hex_selector.GetValue()

    def _update_length(self):
        if self._maximum:
            self._value_control.SetToolTip(f"Minimum characters: {self._minimum}, maximum characters: {self._maximum}")
            self._value_control.SetMaxLength(self._maximum)


class WxPythonBooleanView(WxPythonView, ValueInterface):
    def __init__(self, sizer: wx.Sizer, value_control: wx.CheckBox = None,
                 optional_control: Optional[wx.CheckBox] = None):
        super(WxPythonBooleanView, self).__init__(sizer, optional_control)
        self._value_control = value_control

    def register_change_event(self, callback: Callable):
        # noinspection PyUnusedLocal
        def event_closure(e: wx.Event):
            del e
            callback()

        if self._value_control is not None:
            self._value_control.Bind(wx.EVT_CHECKBOX, event_closure)

    def get_value(self) -> bool:
        return self._value_control.GetValue()

    def set_value(self, val: bool):
        self._value_control.SetValue(val)

    def enable(self, enabled: bool):
        self._value_control.Enable(enabled)


class WxPythonContainerView(WxPythonView, ContainerView):
    def __init__(self, sizer: wx.Sizer, container_sizer: wx.FlexGridSizer, optional_control: Optional[wx.CheckBox] = None):
        super(WxPythonContainerView, self).__init__(sizer, optional_control)
        self._container_sizer = container_sizer
        self._children: List[wx.Sizer] = []

    def add_child(self, view: WxPythonView):
        if self.any_shown(view._sizer):
            self._children.append(view._sizer)
            self._container_sizer.Add(view._sizer)
            if not self.get_has_value():
                self._sizer.Hide(view._sizer, recursive=True)

    def enable(self, enabled: bool):
        if enabled:
            for child in self._children:
                self._sizer.Show(child, recursive=True)
        else:
            for child in self._children:
                self._sizer.Hide(child, recursive=True)
        self._sizer.Layout()
        containing_window = self._sizer.GetContainingWindow()
        if containing_window is not None:
            self._sizer.GetContainingWindow().Layout()

    @staticmethod
    def any_shown(sizer: wx.Sizer) -> bool:
        return any([child.IsShown() for child in sizer.GetChildren()])


class WxPythonListView(WxPythonView, ListView, ValueInterface):
    def __init__(self, sizer: wx.Sizer, num_control: wx.SpinCtrl, optional_control: Optional[wx.CheckBox] = None):
        super(WxPythonListView, self).__init__(sizer, optional_control)
        self._num_control = num_control
        self._children: List[wx.Sizer] = []

    def register_change_event(self, callback: Callable):
        # noinspection PyUnusedLocal
        def event_closure(e: wx.Event):
            del e
            callback()

        self._num_control.Bind(wx.EVT_SPINCTRL, event_closure)

    def get_value(self) -> str:
        return self._num_control.GetValue()

    def set_value(self, val: str):
        self._num_control.SetValue(val)

    def enable(self, enabled: bool):
        self._num_control.Enable(enabled)
        if enabled:
            for child in self._children:
                self._sizer.Show(child, recursive=True)
        else:
            for child in self._children:
                self._sizer.Hide(child, recursive=True)

    def add(self, view: WxPythonView):
        self._sizer.Add(view._sizer)
        self._sizer.Layout()
        self._children.append(view._sizer)

    def remove(self, view: WxPythonView):
        self._children.remove(view._sizer)
        view._sizer.Clear(True)
        self._sizer.Remove(view._sizer)
        self._sizer.Layout()


class WxPythonChoiceView(WxPythonView, ChoiceView, ValueInterface):
    def __init__(self, sizer: wx.Sizer, choice_control: wx.Choice, optional_control: Optional[wx.CheckBox] = None):
        super(WxPythonChoiceView, self).__init__(sizer, optional_control)
        self._choice_control = choice_control
        self._view = None

    def register_change_event(self, callback: Callable):
        # noinspection PyUnusedLocal
        def event_closure(e: wx.Event):
            del e
            callback()

        self._choice_control.Bind(wx.EVT_CHOICE, event_closure)

    def get_value(self) -> str:
        return self._choice_control.GetStringSelection()

    def set_value(self, val: str):
        self._choice_control.SetStringSelection(val)

    def enable(self, enabled: bool):
        self._choice_control.Enable(enabled)
        if self._view is not None:
            self._view.enable(enabled)

    def set_view(self, view: WxPythonView):
        if self._view is not None:
            self._view._sizer.Clear(True)
            self._sizer.Remove(self._view._sizer)
        self._sizer.Add(view._sizer)
        self._view = view
        self._sizer.Layout()

        if not self._choice_control.Enabled:
            self._view.enable(False)


class WxPythonBitstringView(WxPythonView, BitstringInterface):
    def __init__(self, sizer: wx.Sizer, checkboxes: List[Tuple[int, wx.CheckBox]], optional_control: Optional[wx.CheckBox] = None):
        super(WxPythonBitstringView, self).__init__(sizer, optional_control)
        self._checkboxes = checkboxes

    def enable(self, enabled: bool):
        for _, checkbox in self._checkboxes:
            checkbox.Enable(enabled)

    def get_values(self) -> typing.List[int]:
        values = []
        for bit, checkbox in self._checkboxes:
            if checkbox.GetValue():
                values.append(bit)
        return values

    def set_values(self, values: typing.List[int]):
        for bit, checkbox in self._checkboxes:
            checkbox.SetValue(bit in values)
